@@ -449,7 +449,6 @@ static int protolayer_iter_ctx_finish(struct protolayer_iter_ctx *ctx, int ret)
 
 	mm_ctx_delete(&ctx->pool);
 	free(ctx);
-	session2_unhandle(s);
 
 	return ret;
 }
@@ -616,8 +615,6 @@ static int session2_submit(
 {
 	if (session->closing)
 		return kr_error(ECANCELED);
-	if (session->ref_count >= INT_MAX - 1)
-		return kr_error(ETOOMANYREFS);
 	if (kr_fails_assert(session->proto < KR_PROTO_COUNT))
 		return kr_error(EFAULT);
 
@@ -651,7 +648,6 @@ static int session2_submit(
 		.finished_cb = cb,
 		.finished_cb_baton = baton
 	};
-	session->ref_count++;
 	if (had_comm_param) {
 		struct comm_addr_storage *addrst = &ctx->comm_addr_storage;
 		if (comm->src_addr) {
@@ -734,7 +730,7 @@ enum protolayer_iter_cb_result protolayer_break(struct protolayer_iter_ctx *ctx,
 }
 
 
-int wire_buf_init(struct wire_buf *wb, size_t initial_size)
+void wire_buf_init(struct wire_buf *wb, size_t initial_size)
 {
 	char *buf = malloc(initial_size);
 	kr_require(buf);
@@ -743,8 +739,6 @@ int wire_buf_init(struct wire_buf *wb, size_t initial_size)
 		.buf = buf,
 		.size = initial_size
 	};
-
-	return kr_ok();
 }
 
 void wire_buf_deinit(struct wire_buf *wb)
@@ -871,13 +865,12 @@ struct session2 *session2_new(enum session2_transport_type transport_type,
 
 	memcpy(&s->layer_data, offsets, sizeof(offsets));
 	queue_init(s->waiting);
-	int ret = wire_buf_init(&s->wire_buf, wire_buf_length);
-	kr_require(!ret);
+	wire_buf_init(&s->wire_buf, wire_buf_length);
 
-	ret = uv_timer_init(uv_default_loop(), &s->timer);
+	int ret = uv_timer_init(uv_default_loop(), &s->timer);
 	kr_require(!ret);
 	s->timer.data = s;
-	s->ref_count++; /* Session owns the timer */
+	session2_inc_refs(s); /* Session owns the timer */
 
 	/* Initialize the layer's session data */
 	for (size_t i = 0; i < grp->num_layers; i++) {
@@ -900,7 +893,7 @@ struct session2 *session2_new(enum session2_transport_type transport_type,
 
 /** De-allocates the session. Must only be called once the underlying IO handle
  * and timer are already closed, otherwise may leak resources. */
-static void session2_free(struct session2 *s)
+void session2_free(struct session2 *s)
 {
 	const struct protolayer_grp *grp = &protolayer_grps[s->proto];
 	for (size_t i = 0; i < grp->num_layers; i++) {
@@ -917,17 +910,8 @@ static void session2_free(struct session2 *s)
 	free(s);
 }
 
-void session2_unhandle(struct session2 *s)
-{
-	if (kr_fails_assert(s->ref_count > 0)) {
-		session2_free(s);
-		return;
-	}
-
-	s->ref_count--;
-	if (s->ref_count <= 0)
-		session2_free(s);
-}
+extern inline void session2_inc_refs(struct session2 *s);
+extern inline void session2_dec_refs(struct session2 *s);
 
 int session2_start_read(struct session2 *session)
 {
@@ -1110,7 +1094,8 @@ void session2_tasklist_finalize(struct session2 *session, int status)
 		defer_sample_start(&defer_prev_sample_state);
 		do {
 			struct qr_task *t = session2_tasklist_del_first(session, false);
-			kr_require(worker_task_numrefs(t) > 0);
+			if (kr_fails_assert(worker_task_numrefs(t) > 0))
+				continue; // t has been freed already?
 			worker_task_finalize(t, status);
 			worker_task_unref(t);
 			defer_sample_restart();
@@ -1702,14 +1687,14 @@ static inline int session2_transport_push(struct session2 *s,
 static void on_session2_handle_close(uv_handle_t *handle)
 {
 	struct session2 *session = handle->data;
-	kr_require(session->transport.type == SESSION2_TRANSPORT_IO &&
+	kr_assert(session->transport.type == SESSION2_TRANSPORT_IO &&
 			session->transport.io.handle == handle);
 	io_free(handle);
 }
 
 static void on_session2_timer_close(uv_handle_t *handle)
 {
-	session2_unhandle(handle->data);
+	session2_dec_refs(handle->data);
 }
 
 static int session2_handle_close(struct session2 *s)
@@ -1724,8 +1709,8 @@ static int session2_handle_close(struct session2 *s)
 		 * been ended. We do not `uv_close` the handles, we just free
 		 * up the memory. */
 
-		session2_unhandle(s); /* For timer handle */
-		io_free(handle); /* This will unhandle the transport handle */
+		session2_dec_refs(s); /* For timer handle */
+		io_free(handle); /* This will decrement refs for the transport handle */
 		return kr_ok();
 	}
 
@@ -1751,7 +1736,10 @@ static int session2_transport_event(struct session2 *s,
 	bool is_close_event = (event == PROTOLAYER_EVENT_CLOSE ||
 			event == PROTOLAYER_EVENT_FORCE_CLOSE);
 	if (is_close_event) {
-		kr_require(session2_is_empty(s));
+		if (kr_fails_assert(session2_is_empty(s))) {
+			session2_waitinglist_finalize(s, KR_STATE_FAIL);
+			session2_tasklist_finalize(s, KR_STATE_FAIL);
+		}
 		session2_timer_stop(s);
 		s->closing = true;
 	}
