@@ -130,13 +130,21 @@ class Server:
                     f"Configuration file was not found at '{self._config_path}'."
                     " Something must have happened to it while we were running."
                 )
-                logger.error("Configuration have NOT been changed.")
+                logger.error("Configuration has NOT been changed.")
             except (DataParsingError, DataValidationError) as e:
                 logger.error(f"Failed to parse the updated configuration file: {e}")
-                logger.error("Configuration have NOT been changed.")
+                logger.error("Configuration has NOT been changed.")
             except KresManagerException as e:
                 logger.error(f"Reloading of the configuration file failed: {e}")
-                logger.error("Configuration have NOT been changed.")
+                logger.error("Configuration has NOT been changed.")
+
+    async def _renew_config(self) -> None:
+        try:
+            await self.config_store.renew()
+            logger.info("Configuration successfully renewed")
+        except KresManagerException as e:
+            logger.error(f"Renewing the configuration failed: {e}")
+            logger.error("Configuration has NOT been renewed.")
 
     async def sigint_handler(self) -> None:
         logger.info("Received SIGINT, triggering graceful shutdown")
@@ -325,6 +333,15 @@ class Server:
         await self._reload_config()
         return web.Response(text="Reloading...")
 
+    async def _handler_renew(self, _request: web.Request) -> web.Response:
+        """
+        Route handler for renewing the configuration
+        """
+
+        logger.info("Renewing configuration event triggered...")
+        await self._renew_config()
+        return web.Response(text="Renewing configuration...")
+
     async def _handler_processes(self, request: web.Request) -> web.Response:
         """
         Route handler for listing PIDs of subprocesses
@@ -359,6 +376,7 @@ class Server:
                 web.patch(r"/v1/config{path:.*}", self._handler_config_query),
                 web.post("/stop", self._handler_stop),
                 web.post("/reload", self._handler_reload),
+                web.post("/renew", self._handler_renew),
                 web.get("/schema", self._handler_schema),
                 web.get("/schema/ui", self._handle_view_schema),
                 web.get("/metrics", self._handler_metrics),
