@@ -2,15 +2,15 @@
 
 set -e
 
-curl --silent --fail --unix-socket /var/run/knot-resolver/manager.sock http://localhost/metrics > /dev/null
+curl --silent --fail --unix-socket /run/knot-resolver/kres-api.sock http://localhost/metrics > /dev/null
 
-kresctl metrics
+kresctl metrics > /dev/null
 if [ "$?" -ne "0" ]; then
 	echo "Could not get metrics in JSON format"
 	exit 1
 fi
 
-kresctl metrics --prometheus
+kresctl metrics --prometheus > /dev/null
 if [ "$?" -ne "0" ]; then
 	echo "Could not get metrics in Prometheus format"
 	exit 1
